#pragma once

#include <functional>
#include <list>
#include <memory>
#include <vector>

#include <dune/common/std/optional.hh>

#include <amdis/Boundary.hpp>
#include <amdis/BoundaryCondition.hpp>
#include <amdis/common/Concepts.hpp>
#include <amdis/utility/RangeType.hpp>
#include <amdis/utility/TreeData.hpp>
#include <amdis/utility/Visitor.hpp>

namespace AMDiS
{
  /// Implements a boundary condition of Dirichlet-type.
  /**
   * By calling the methods \ref init() and \ref finish before and after
   * assembling the system-matrix, respectively, dirichlet boundary conditions
   * can be applied to the matrix and system vector. Therefore, a predicate
   * functions indicates the DOFs where values should be enforced and a second
   * functor provided in the constructor is responsible for determining the
   * values to be set at the DOFs.
   *
   * In the \ref finish method the matrix is called with \ref applyDirichletBC
   * to erase the corresponding rows and columns for the DOF indices. This
   * application of boundary conditions can be symmetric if the matrix does
   * support this symmetric modification. As a result, this method returns a list
   * of columns values, that should be subtracted from the rhs.
   **/
  template <class Domain, class Range = double>
  class DirichletBC
      : public BoundaryCondition
  {
    using Super = BoundaryCondition;

  public:
    template <class BM, class Values,
      REQUIRES(Concepts::Functor<Values, Range(Domain)>) >
    DirichletBC(BM&& boundaryManager, BoundaryType id, Values&& values)
      : Super(std::forward<BM>(boundaryManager), id)
      , values_(std::forward<Values>(values))
    {}

    template <class Predicate, class Values,
      REQUIRES(Concepts::Functor<Predicate, bool(Domain)>),
      REQUIRES(Concepts::Functor<Values, Range(Domain)>)>
    DirichletBC(Predicate&& predicate, Values&& values)
      : predicate_(std::forward<Predicate>(predicate))
      , values_(std::forward<Values>(values))
    {}

    template <class Intersection>
    bool onBoundary(Intersection const& intersection) const
    {
      return predicate_ ? (*predicate_)(intersection.geometry().center())
                        : Super::onBoundary(intersection);
    }

    // fill \ref dirichletNodes_ with 1 or 0 whether DOF corresponds to the boundary or not.
    template <class RowBasis, class ColBasis>
    void init(RowBasis const& rowBasis, ColBasis const& colBasis);

    /// \brief Apply dirichlet BC to matrix and vector, i.e., add a unit-row
    /// to the matrix and optionally delete the corresponding matrix-column.
    template <class Matrix, class VectorX, class VectorB, class RowNode, class ColNode>
    void fillBoundaryCondition(Matrix& matrix, VectorX& solution, VectorB& rhs,
                               RowNode const& rowNode, ColNode const& colNode);

  private:
    Dune::Std::optional<std::function<bool(Domain)>> predicate_;
    std::function<Range(Domain)> values_;
    std::vector<bool> dirichletNodes_;
  };


  template <class Basis>
  struct DirichletData
  {
    using WorldVector = typename Basis::GridView::template Codim<0>::Geometry::GlobalCoordinate;

    template <class RowNode, class ColNode>
    using type = std::list< std::shared_ptr<DirichletBC<WorldVector, RangeType_t<RowNode>>> >;
  };

  template <class RowBasis, class ColBasis>
  using DirichletBCs = MatrixData<RowBasis, ColBasis, DirichletData<RowBasis>::template type>;

} // end namespace AMDiS

#include "DirichletBC.inc.hpp"
