#pragma once

#include <functional>
#include <memory>
#include <type_traits>

#include <dune/common/shared_ptr.hh>
#include <dune/geometry/quadraturerules.hh>

#include <amdis/ContextGeometry.hpp>
#include <amdis/LocalAssemblerBase.hpp>
#include <amdis/common/Concepts.hpp>
#include <amdis/utility/FiniteElementType.hpp>

namespace AMDiS
{
  /// Implementation of interface \ref LocalAssemblerBase
  template <class LocalContext, class Operator, class... Nodes>
  class LocalAssembler
      : public LocalAssemblerBase<LocalContext, Nodes...>
  {
    using Super = LocalAssemblerBase<LocalContext, Nodes...>;

  private:

    using Element = typename Super::Element;
    using Geometry = typename Super::Geometry;
    using ElementMatrixVector = typename Super::ElementMatrixVector;

  public:

    /// Constructor. Stores a copy of operator `op`.
    explicit LocalAssembler(Operator const& op)
      : op_(Dune::wrap_or_move(op))
    {}

    /// Constructor. Stores a copy of operator `op`.
    explicit LocalAssembler(Operator&& op)
      : op_(Dune::wrap_or_move(std::move(op)))
    {}

    /// Constructor. Stores the reference to the operator.
    explicit LocalAssembler(std::reference_wrapper<Operator> op)
      : op_(Dune::wrap_or_move(op.get()))
    {}

    /// \brief Implementation of \ref LocalAssemblerBase::bind.
    /**
     * Binds the operator `op_` to the `element` and `geometry` and
     * stores point to the `element` and `geometry`.
     **/
    virtual void bind(Element const& element, Geometry const& geometry) final
    {
      element_ = &element;
      geometry_ = &geometry;
      op_->bind(element, geometry);
    }

    /// \brief Implementation of \ref LocalAssemblerBase::unbind
    /**
     * Unbinds the operator `op_` and sets \ref element_ and \ref geometry_
     * to nullptr.
     **/
    virtual void unbind() final
    {
      op_->unbind();
      geometry_ = nullptr;
      element_ = nullptr;
    }

    /// Implementation of \ref LocalAssemblerBase::assemble
    /**
     * Stores geometry and localGeometry and calls
     * \ref calculateElementVector or \ref calculateElementMatrix on the
     * vector or matrix operator, respectively.
     **/
    virtual void assemble(LocalContext const& localContext,
                          Nodes const&... nodes,
                          ElementMatrixVector& elementMatrixVector) final
    {
      ContextGeometry<LocalContext> context{localContext, element(), geometry()};
      assembleImpl(context, nodes..., elementMatrixVector);
    }


#ifndef DOXYGEN

  protected: // implementation detail

    // matrix assembling
    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void assembleImpl(Context const& context,
                      RowNode const& rowNode, ColNode const& colNode,
                      ElementMatrix& elementMatrix)
    {
      op_->calculateElementMatrix(context, rowNode, colNode, elementMatrix);
    }

    // vector assembling
    template <class Context, class Node, class ElementVector>
    void assembleImpl(Context const& context,
                      Node const& node,
                      ElementVector& elementVector)
    {
      op_->calculateElementVector(context, node, elementVector);
    }

#endif // DOXYGEN

  public:

    /// return the bound entity (of codim 0)
    Element const& element() const
    {
      assert( element_ );
      return *element_;
    }

    /// return the geometry of the bound element
    Geometry const& geometry() const
    {
      assert( geometry_ );
      return *geometry_;
    }


  private:

    /// the stored operator, implementing \ref GridFunctionOperatorBase
    std::shared_ptr<Operator> op_;

    Element const* element_ = nullptr;
    Geometry const* geometry_ = nullptr;
  };


  /// Generate a \ref LocalAssembler on a given `LocalContext` (element or intersection)
  template <class LocalContext, class Operator, class... Nodes>
  auto makeLocalAssembler(Operator&& op, Nodes const&...)
  {
    return LocalAssembler<LocalContext, Underlying_t<Operator>, Nodes...>{std::forward<Operator>(op)};
  }

} // end namespace AMDiS
