#pragma once

#include <type_traits>
#include <vector>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/LocalBasisCache.hpp>
#include <amdis/common/Size.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct stokes {};
  }


  /// Stokes operator \f$ \langle\nabla\mathbf{v}, \nu \nabla\mathbf{u}\rangle + \langle\nabla\cdot\mathbf{v}, p\rangle + \langle q, \langle\nabla\cdot\mathbf{u}\rangle \f$
  template <class LocalContext, class ViscosityExpr>
  class GridFunctionOperator<tag::stokes, LocalContext, ViscosityExpr>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::stokes, LocalContext, ViscosityExpr>,
                                        LocalContext, ViscosityExpr>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, ViscosityExpr>;

    static_assert( Size_v<typename ViscosityExpr::Range> == 1, "Viscosity must be of scalar type." );

  public:
    GridFunctionOperator(tag::stokes, ViscosityExpr const& expr)
      : Super(expr, 1)
    {}

    template <class Context, class Node, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          Node const& tree,
                          Node const& /*colNode*/,
                          ElementMatrix& elementMatrix)
    {
      using VelocityNode = typename Node::template Child<0>::type;
      using VelocityCompNode = typename VelocityNode::template Child<0>::type;
      using PressureNode = typename Node::template Child<1>::type;
      static_assert(VelocityNode::isPower && PressureNode::isLeaf, "");

      using namespace Dune::Indices;

      auto const& quad = this->getQuadratureRule(context.type(), tree, tree);

      auto const& velocityLocalFE = tree.child(_0,0).finiteElement();
      std::size_t numVelocityLocalFE = velocityLocalFE.size();

      auto const& pressureLocalFE = tree.child(_1).finiteElement();
      std::size_t numPressureLocalFE = pressureLocalFE.size();

      NodeQuadCache<VelocityCompNode> velocityCache(velocityLocalFE.localBasis());
      NodeQuadCache<PressureNode> pressureCache(pressureLocalFE.localBasis());

      auto const& shapeGradientsCache = velocityCache.evaluateJacobianAtQP(context, quad);
      auto const& pressureValuesCache = pressureCache.evaluateFunctionAtQP(context, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(quad[iq].position()) * quad[iq].weight();
        const auto vel_factor = Super::coefficient(local) * factor;

        auto const& shapeGradients = shapeGradientsCache[iq];

        // Compute the shape function gradients on the real element
        using RangeFieldType = typename NodeQuadCache<VelocityCompNode>::Traits::RangeFieldType;
        using WorldVector = Dune::FieldVector<RangeFieldType,Context::dow>;
        thread_local std::vector<WorldVector> gradients;
        gradients.resize(shapeGradients.size());
        for (std::size_t i = 0; i < gradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], gradients[i]);

        auto const& pressureValues = pressureValuesCache[iq];

        // <viscosity * grad(u), grad(v)>
        for (std::size_t i = 0; i < numVelocityLocalFE; ++i) {
          const auto value_ii = vel_factor * (gradients[i] * gradients[i]);
          for (std::size_t k = 0; k < Context::dow; ++k) {
            const auto local_ki = tree.child(_0,k).localIndex(i);
            elementMatrix[local_ki][local_ki] += value_ii;
          }

          for (std::size_t j = i+1; j < numVelocityLocalFE; ++j) {
            const auto value = vel_factor * (gradients[i] * gradients[j]);
            for (std::size_t k = 0; k < Context::dow; ++k) {
              const auto local_ki = tree.child(_0,k).localIndex(i);
              const auto local_kj = tree.child(_0,k).localIndex(j);
              elementMatrix[local_ki][local_kj] += value;
              elementMatrix[local_kj][local_ki] += value;
            }
          }
        }

        // <p, div(v)> + <div(u), q>
        for (std::size_t i = 0; i < numVelocityLocalFE; ++i) {
          for (std::size_t j = 0; j < numPressureLocalFE; ++j) {
            const auto value = factor * pressureValues[j];
            for (std::size_t k = 0; k < Context::dow; ++k) {
              const auto local_v = tree.child(_0,k).localIndex(i);
              const auto local_p = tree.child(_1).localIndex(j);

              elementMatrix[local_v][local_p] += gradients[i][k] * value;
              elementMatrix[local_p][local_v] += gradients[i][k] * value;
            }
          }
        }
      }

    } // end calculateElementMatrix
  };

  /** @} **/

} // end namespace AMDiS
