#pragma once

#include <cmath>
#include <utility>

#include <dune/functions/functionspacebases/sizeinfo.hh>

#include <amdis/DataTransfer.hpp>
#include <amdis/GridTransferManager.hpp>
#include <amdis/LocalAssemblerList.hpp>
#include <amdis/common/Math.hpp>
#include <amdis/common/ScalarTypes.hpp>
#include <amdis/linear_algebra/DOFVectorInterface.hpp>
#include <amdis/utility/MultiIndex.hpp>
#include <amdis/utility/TreePath.hpp>

namespace AMDiS
{
  /// The basic container that stores a base vector and a corresponding basis
  template <class BasisType, class Backend>
  class DOFVectorBase
      : public DOFVectorInterface
  {
    using Self = DOFVectorBase;

  public:
    /// The type of the functionspace basis associated to this vector
    using Basis = BasisType;
    using LocalView = typename Basis::LocalView;

    using Element = typename LocalView::Element;
    using Geometry = typename Element::Geometry;

    /// The index/size - type
    using size_type  = typename BasisType::size_type;

    /// The type of the elements of the DOFVector
    using value_type = typename Backend::value_type;

    /// The type of the data vector used in the backend
    using BaseVector = typename Backend::BaseVector;
    using ElementVector = Dune::DynamicVector<double>;

    /// Defines an interface to transfer the data during grid adaption
    using DataTransfer = DataTransferInterface<Self>;

    /// A creator for a concrete data transfer object, depending on \ref DataTransferOperation
    using DataTransferFactory = AMDiS::DataTransferFactory<Self>;

  public:
    /// Constructor. Constructs new BaseVector.
    DOFVectorBase(BasisType const& basis, DataTransferOperation op = NO_OPERATION)
      : basis_(&basis)
      , dataTransfer_(DataTransferFactory::create(op, basis))
    {
      compress();
      GridTransferManager::attach(*this);
      operators_.init(basis);
    }

    /// Copy constructor
    DOFVectorBase(Self const& that)
      : basis_(that.basis_)
      , backend_(that.backend_)
      , elementVector_(that.elementVector_)
      , operators_(that.operators_)
      , dataTransfer_(that.dataTransfer_)
    {
      GridTransferManager::attach(*this);
    }

    /// Move constructor
    DOFVectorBase(Self&& that)
      : basis_(std::move(that.basis_))
      , backend_(std::move(that.backend_))
      , elementVector_(std::move(that.elementVector_))
      , operators_(std::move(that.operators_))
      , dataTransfer_(std::move(that.dataTransfer_))
    {
      GridTransferManager::attach(*this);
    }

    /// Destructor
    virtual ~DOFVectorBase() override
    {
      GridTransferManager::detach(*this);
    }

    /// Copy assignment operator
    Self& operator=(Self const& that)
    {
      GridTransferManager::detach(*this);
      basis_ = that.basis_;
      backend_.resize(that.size());
      backend_ = that.backend_;
      dataTransfer_ = that.dataTransfer_;
      GridTransferManager::attach(*this);
      return *this;
    }

    /// Move assignment
    Self& operator=(Self&& that) = default;

    /// Sets each DOFVector to the scalar \p value.
    template <class Scalar,
      std::enable_if_t<Concepts::Arithmetic<Scalar>, int> = 0>
    Self& operator=(Scalar value)
    {
      backend_.set(value);
      return *this;
    }

    /// Return the basis \ref basis_ associated to the vector
    Basis const& basis() const
    {
      return *basis_;
    }

    /// Return the data-vector
    BaseVector const& vector() const
    {
      return backend_.vector();
    }

    /// Return the data-vector
    BaseVector& vector()
    {
      return backend_.vector();
    }

    /// Return the size of the \ref basis
    size_type size() const
    {
      return basis_->dimension();
    }

    void resize(Dune::Functions::SizeInfo<Basis> const& s)
    {
      backend_.resize(size_type(s));
    }

    /// Resize the \ref vector to the size of the \ref basis and set to zero
    virtual void compress() override
    {
      if (size_type(backend_.size()) != size()) {
        backend_.resize(size());
        backend_.set(0);
      }
    }

    /// Access the entry \p idx of the \ref vector with read-access.
    template <class Index>
    auto const& operator[](Index idx) const
    {
      size_type i = flatMultiIndex(idx);
      return backend_[i];
    }

    /// Access the entry \p idx of the \ref vector with write-access.
    template <class Index>
    auto& operator[](Index idx)
    {
      size_type i = flatMultiIndex(idx);
      return backend_[i];
    }

    void init(bool asmVector);

    void finish(bool asmVector);

    /// Insert a block of values into the matrix (add to existing values)
    void insert(LocalView const& localView, ElementVector const& elementVector);

    /// Associate a local operator with this DOFVector
    template <class ContextTag, class Operator, class TreePath = RootTreePath>
    void addOperator(ContextTag contextTag, Operator const& preOp, TreePath path = {});

    /// Assemble the vector operators on the bound element.
    void assemble(LocalView const& localView);

    /// Assemble all vector operators
    void assemble();

    /// Return the associated DataTransfer object
    DataTransfer const& dataTransfer() const
    {
      return *dataTransfer_;
    }

    /// Return the associated DataTransfer object
    DataTransfer& dataTransfer()
    {
      return *dataTransfer_;
    }

    /// Create a new DataTransfer object based on the operation type
    void setDataTransfer(DataTransferOperation op)
    {
      dataTransfer_ = DataTransferFactory::create(op, this->basis());
    }

    /// Assign the DataTransfer object
    void setDataTransfer(std::shared_ptr<DataTransfer> const& dataTransfer)
    {
      dataTransfer_ = dataTransfer;
    }

    /// Implementation of \ref DOFVectorInterface::preAdapt
    virtual void preAdapt(bool mightCoarsen) override
    {
      dataTransfer_->preAdapt(*this, mightCoarsen);
    }

    /// Implementation of \ref DOFVectorInterface::postAdapt
    virtual void postAdapt(bool refined) override
    {
      dataTransfer_->postAdapt(*this, refined);
    }

  private:
    /// The finite element space / basis associated with the data vector
    Basis const* basis_;

    /// Data backend
    Backend backend_;

    /// Dense vector to store coefficients during \ref assemble()
    ElementVector elementVector_;

    /// List of operators associated to nodes
    VectorOperators<Basis> operators_;

    /// Data interpolation when the grid changes
    std::shared_ptr<DataTransfer> dataTransfer_;
  };

} // end namespace AMDiS

#include "DOFVectorBase.inc.hpp"
