#pragma once

#include <list>

#include <boost/numeric/mtl/matrix/compressed2D.hpp>
#include <boost/numeric/mtl/matrix/inserter.hpp>
#include <boost/numeric/mtl/utility/property_map.hpp>
#include <boost/numeric/mtl/utility/range_wrapper.hpp>

#include <amdis/linear_algebra/Common.hpp>
#include <amdis/linear_algebra/Constraints.hpp>

namespace AMDiS
{
  template <class T, class P>
  struct Constraints<mtl::compressed2D<T,P>>
  {
    using Matrix = mtl::compressed2D<T,P>;

#ifdef SYMMETRIC_DIRICHLET_BC
    template <class BitVector>
    static std::vector<Triplet<T>> dirichletBC(Matrix& mat, BitVector const& nodes, bool setDiagonal = true)
    {
      std::vector<Triplet<T>> columns;
      if (setDiagonal)
        columns.reserve(std::size_t(mat.nnz()/(0.84*num_rows(mat))));

      // Define the property maps
      auto row   = mtl::mat::row_map(mat);
      auto col   = mtl::mat::col_map(mat);
      auto value = mtl::mat::value_map(mat);

      // iterate over the matrix
      std::size_t slotSize = 0;
      for (auto r : mtl::rows_of(mat)) {   // rows or columns
        std::size_t rowSize = 0;
        for (auto i : mtl::nz_of(r)) {          // non-zeros within
          ++rowSize;
          if (nodes[row(i)]) {
            // set identity row
            value(i, T(0));
          }
          else if (setDiagonal && nodes[col(i)]) {
            columns.push_back({row(i), col(i), value(i)});
            value(i, T(0));
          }
        }
        slotSize = std::max(slotSize, rowSize);
      }

      // set diagonal entry
      if (setDiagonal) {
        mtl::mat::inserter<Matrix, mtl::update_store<T> > ins(mat, slotSize);
        for (std::size_t i = 0; i < nodes.size(); ++i) {
          if (nodes[i])
            ins[i][i] = T(1);
        }
      }

      return columns;
    }
#else
    template <class BitVector>
    static std::vector<Triplet<T>> dirichletBC(Matrix& mat, BitVector const& nodes, bool setDiagonal = true)
    {
      // Define the property maps
      auto row   = mtl::mat::row_map(mat);
      auto col   = mtl::mat::col_map(mat);
      auto value = mtl::mat::value_map(mat);

      // iterate over the matrix
      for (auto r : mtl::rows_of(mat)) {   // rows of the matrix
        if (nodes[r.value()]) {
          for (auto i : mtl::nz_of(r)) {          // non-zeros within
            // set identity row
            value(i, (setDiagonal && row(i) == col(i) ? T(1) : T(0)) );
          }
        }
      }

      return {};
    }
#endif


    template <class Associations>
    static std::size_t at(Associations const& m, std::size_t idx)
    {
      auto it = m.find(idx);
      assert(it != m.end());
      return it->second;
    }

#ifdef SYMMETRIC_PERIODIC_BC
    template <class BitVector, class Associations>
    static std::vector<Triplet<T>> periodicBC(Matrix& mat, BitVector const& left, Associations const& left2right,
      bool setDiagonal = true)
    {
      error_exit("Not implemented");
    }
#else
    template <class BitVector, class Associations>
    static std::vector<Triplet<T>> periodicBC(Matrix& mat, BitVector const& left, Associations const& left2right,
      bool setDiagonal = true)
    {
      std::vector<Triplet<T>> rowValues;
      rowValues.reserve(left2right.size()*std::size_t(mat.nnz()/(0.9*num_rows(mat))));

      // Define the property maps
      // auto row   = mtl::mat::row_map(mat);
      auto col   = mtl::mat::col_map(mat);
      auto value = mtl::mat::value_map(mat);

      // iterate over the matrix
      std::size_t slotSize = 0;
      for (auto r : mtl::rows_of(mat)) {
        if (left[r.value()]) {
          slotSize = std::max(slotSize, std::size_t(mat.nnz_local(r.value())));
          std::size_t right = at(left2right,r.value());

          for (auto i : mtl::nz_of(r)) {
            rowValues.push_back({right,col(i),value(i)});
            value(i, T(0));
          }
        }
      }

      mtl::mat::inserter<Matrix, mtl::update_plus<T> > ins(mat, 2*slotSize);
      for (auto const& t : rowValues)
        ins[t.row][t.col] += t.value;

      if (setDiagonal) {
        for (std::size_t i = 0; i < size(left); ++i) {
          if (left[i]) {
            ins[i][i] = T(1);
            ins[i][at(left2right,i)] = T(-1);
          }
        }
      }

      return {};
    }
#endif
  };

} // end namespace AMDiS
